@@ -1,12 +1,18 @@
+import sys
+import types
 from array import array
 from collections import abc
-import sys
 
 from ._abc import MultiMapping, MutableMultiMapping
 
-
 _marker = object()
 
+if sys.version_info >= (3, 9):
+    GenericAlias = types.GenericAlias
+else:
+    def GenericAlias(cls):
+        return cls
+
 
 class istr(str):
 
@@ -14,21 +20,6 @@ class istr(str):
 
     __is_istr__ = True
 
-    def __new__(cls, val='',
-                encoding=sys.getdefaultencoding(), errors='strict'):
-        if getattr(val, '__is_istr__', False):
-            # Faster than instance check
-            return val
-        if type(val) is str:
-            pass
-        else:
-            val = str(val)
-        val = val.title()
-        return str.__new__(cls, val)
-
-    def title(self):
-        return self
-
 
 upstr = istr  # for relaxing backward compatibility problems
 
@@ -39,11 +30,11 @@ def getversion(md):
     return md._impl._version
 
 
-_version = array('Q', [0])
+_version = array("Q", [0])
 
 
 class _Impl:
-    __slots__ = ('_items', '_version')
+    __slots__ = ("_items", "_version")
 
     def __init__(self):
         self._items = []
@@ -55,9 +46,13 @@ class _Impl:
         v[0] += 1
         self._version = v[0]
 
+    if sys.implementation.name != "pypy":
+
+        def __sizeof__(self):
+            return object.__sizeof__(self) + sys.getsizeof(self._items)
+
 
 class _Base:
-
     def _title(self, key):
         return key
 
@@ -69,7 +64,7 @@ class _Base:
             return res
         if not res and default is not _marker:
             return default
-        raise KeyError('Key not found: %r' % key)
+        raise KeyError("Key not found: %r" % key)
 
     def getone(self, key, default=_marker):
         """Get first value matching the key."""
@@ -79,7 +74,7 @@ class _Base:
                 return v
         if default is not _marker:
             return default
-        raise KeyError('Key not found: %r' % key)
+        raise KeyError("Key not found: %r" % key)
 
     # Mapping interface #
 
@@ -87,10 +82,7 @@ class _Base:
         return self.getone(key)
 
     def get(self, key, default=None):
-        """Get first value matching the key.
-
-        The method is alias for .getone().
-        """
+        """Get first value matching the key."""
         return self.getone(key, default)
 
     def __iter__(self):
@@ -139,24 +131,26 @@ class _Base:
         return False
 
     def __repr__(self):
-        body = ', '.join("'{}': {!r}".format(k, v) for k, v in self.items())
-        return '<{}({})>'.format(self.__class__.__name__, body)
+        body = ", ".join("'{}': {!r}".format(k, v) for k, v in self.items())
+        return "<{}({})>".format(self.__class__.__name__, body)
+
+    __class_getitem__ = classmethod(GenericAlias)
 
 
 class MultiDictProxy(_Base, MultiMapping):
+    """Read-only proxy for MultiDict instance."""
 
     def __init__(self, arg):
         if not isinstance(arg, (MultiDict, MultiDictProxy)):
             raise TypeError(
-                'ctor requires MultiDict or MultiDictProxy instance'
-                ', not {}'.format(
-                    type(arg)))
+                "ctor requires MultiDict or MultiDictProxy instance"
+                ", not {}".format(type(arg))
+            )
 
         self._impl = arg._impl
 
     def __reduce__(self):
-        raise TypeError("can't pickle {} objects".format(
-            self.__class__.__name__))
+        raise TypeError("can't pickle {} objects".format(self.__class__.__name__))
 
     def copy(self):
         """Return a copy of itself."""
@@ -164,13 +158,14 @@ class MultiDictProxy(_Base, MultiMapping):
 
 
 class CIMultiDictProxy(MultiDictProxy):
+    """Read-only proxy for CIMultiDict instance."""
 
     def __init__(self, arg):
         if not isinstance(arg, (CIMultiDict, CIMultiDictProxy)):
             raise TypeError(
-                'ctor requires CIMultiDict or CIMultiDictProxy instance'
-                ', not {}'.format(
-                    type(arg)))
+                "ctor requires CIMultiDict or CIMultiDictProxy instance"
+                ", not {}".format(type(arg))
+            )
 
         self._impl = arg._impl
 
@@ -183,12 +178,17 @@ class CIMultiDictProxy(MultiDictProxy):
 
 
 class MultiDict(_Base, MutableMultiMapping):
+    """Dictionary with the support for duplicate keys."""
 
     def __init__(self, *args, **kwargs):
         self._impl = _Impl()
 
-        self._extend(args, kwargs, self.__class__.__name__,
-                     self._extend_items)
+        self._extend(args, kwargs, self.__class__.__name__, self._extend_items)
+
+    if sys.implementation.name != "pypy":
+
+        def __sizeof__(self):
+            return object.__sizeof__(self) + sys.getsizeof(self._impl)
 
     def __reduce__(self):
         return (self.__class__, (list(self.items()),))
@@ -198,10 +198,11 @@ class MultiDict(_Base, MutableMultiMapping):
 
     def _key(self, key):
         if isinstance(key, str):
-            return str(key)
+            return key
         else:
-            raise TypeError("MultiDict keys should be either str "
-                            "or subclasses of str")
+            raise TypeError(
+                "MultiDict keys should be either str " "or subclasses of str"
+            )
 
     def add(self, key, value):
         identity = self._title(key)
@@ -213,38 +214,48 @@ class MultiDict(_Base, MutableMultiMapping):
         cls = self.__class__
         return cls(self.items())
 
+    __copy__ = copy
+
     def extend(self, *args, **kwargs):
         """Extend current MultiDict with more values.
 
         This method must be used instead of update.
         """
-        self._extend(args, kwargs, 'extend', self._extend_items)
+        self._extend(args, kwargs, "extend", self._extend_items)
 
     def _extend(self, args, kwargs, name, method):
         if len(args) > 1:
-            raise TypeError("{} takes at most 1 positional argument"
-                            " ({} given)".format(name, len(args)))
+            raise TypeError(
+                "{} takes at most 1 positional argument"
+                " ({} given)".format(name, len(args))
+            )
         if args:
             arg = args[0]
-            if isinstance(args[0], (MultiDict, MultiDictProxy)):
+            if isinstance(args[0], (MultiDict, MultiDictProxy)) and not kwargs:
                 items = arg._impl._items
             else:
-                if hasattr(arg, 'items'):
+                if hasattr(arg, "items"):
                     arg = arg.items()
+                if kwargs:
+                    arg = list(arg)
+                    arg.extend(list(kwargs.items()))
                 items = []
                 for item in arg:
                     if not len(item) == 2:
                         raise TypeError(
                             "{} takes either dict or list of (key, value) "
-                            "tuples".format(name))
-                    items.append((self._title(item[0]),
-                                  self._key(item[0]),
-                                  item[1]))
+                            "tuples".format(name)
+                        )
+                    items.append((self._title(item[0]), self._key(item[0]), item[1]))
 
             method(items)
-
-        method([(self._title(key), key, value)
-               for key, value in kwargs.items()])
+        else:
+            method(
+                [
+                    (self._title(key), self._key(key), value)
+                    for key, value in kwargs.items()
+                ]
+            )
 
     def _extend_items(self, items):
         for identity, key, value in items:
@@ -258,15 +269,14 @@ class MultiDict(_Base, MutableMultiMapping):
     # Mapping interface #
 
     def __setitem__(self, key, value):
-        key = self._title(key)
         self._replace(key, value)
 
     def __delitem__(self, key):
-        key = self._title(key)
+        identity = self._title(key)
         items = self._impl._items
         found = False
         for i in range(len(items) - 1, -1, -1):
-            if items[i][0] == key:
+            if items[i][0] == identity:
                 del items[i]
                 found = True
         if not found:
@@ -276,9 +286,9 @@ class MultiDict(_Base, MutableMultiMapping):
 
     def setdefault(self, key, default=None):
         """Return value for key, set value to default if key is not present."""
-        key = self._title(key)
+        identity = self._title(key)
         for i, k, v in self._impl._items:
-            if i == key:
+            if i == identity:
                 return v
         self.add(key, default)
         return default
@@ -290,9 +300,9 @@ class MultiDict(_Base, MutableMultiMapping):
         KeyError is raised.
 
         """
-        key = self._title(key)
+        identity = self._title(key)
         for i in range(len(self._impl._items)):
-            if self._impl._items[i][0] == key:
+            if self._impl._items[i][0] == identity:
                 value = self._impl._items[i][2]
                 del self._impl._items[i]
                 self._impl.incr_version()
@@ -302,7 +312,7 @@ class MultiDict(_Base, MutableMultiMapping):
         else:
             return default
 
-    pop = popone
+    pop = popone  # type: ignore
 
     def popall(self, key, default=_marker):
         """Remove all occurrences of key and return the list of corresponding
@@ -315,7 +325,7 @@ class MultiDict(_Base, MutableMultiMapping):
         found = False
         identity = self._title(key)
         ret = []
-        for i in range(len(self._impl._items)-1, -1, -1):
+        for i in range(len(self._impl._items) - 1, -1, -1):
             item = self._impl._items[i]
             if item[0] == identity:
                 ret.append(item[2])
@@ -342,7 +352,7 @@ class MultiDict(_Base, MutableMultiMapping):
 
     def update(self, *args, **kwargs):
         """Update the dictionary from *other*, overwriting existing keys."""
-        self._extend(args, kwargs, 'update', self._update_items)
+        self._extend(args, kwargs, "update", self._update_items)
 
     def _update_items(self, items):
         if not items:
@@ -381,7 +391,7 @@ class MultiDict(_Base, MutableMultiMapping):
         identity = self._title(key)
         items = self._impl._items
 
-        for i in range(len(items)-1, -1, -1):
+        for i in range(len(items)):
             item = items[i]
             if item[0] == identity:
                 items[i] = (identity, key, value)
@@ -394,35 +404,49 @@ class MultiDict(_Base, MutableMultiMapping):
             self._impl.incr_version()
             return
 
-        # remove all precending items
-        i = 0
-        while i < rgt:
+        # remove all tail items
+        i = rgt + 1
+        while i < len(items):
             item = items[i]
             if item[0] == identity:
                 del items[i]
-                rgt -= 1
             else:
                 i += 1
 
 
 class CIMultiDict(MultiDict):
+    """Dictionary with the support for duplicate case-insensitive keys."""
 
     def _title(self, key):
         return key.title()
 
 
-class _ViewBase:
+class _Iter:
+    __slots__ = ("_size", "_iter")
 
+    def __init__(self, size, iterator):
+        self._size = size
+        self._iter = iterator
+
+    def __iter__(self):
+        return self
+
+    def __next__(self):
+        return next(self._iter)
+
+    def __length_hint__(self):
+        return self._size
+
+
+class _ViewBase:
     def __init__(self, impl):
         self._impl = impl
-        self._version = impl._version
 
     def __len__(self):
         return len(self._impl._items)
 
 
 class _ItemsView(_ViewBase, abc.ItemsView):
-
     def __contains__(self, item):
         assert isinstance(item, tuple) or isinstance(item, list)
         assert len(item) == 2
@@ -432,8 +456,11 @@ class _ItemsView(_ViewBase, abc.ItemsView):
         return False
 
     def __iter__(self):
+        return _Iter(len(self), self._iter(self._impl._version))
+
+    def _iter(self, version):
         for i, k, v in self._impl._items:
-            if self._version != self._impl._version:
+            if version != self._impl._version:
                 raise RuntimeError("Dictionary changed during iteration")
             yield k, v
 
@@ -441,12 +468,11 @@ class _ItemsView(_ViewBase, abc.ItemsView):
         lst = []
         for item in self._impl._items:
             lst.append("{!r}: {!r}".format(item[1], item[2]))
-        body = ', '.join(lst)
-        return '{}({})'.format(self.__class__.__name__, body)
+        body = ", ".join(lst)
+        return "{}({})".format(self.__class__.__name__, body)
 
 
 class _ValuesView(_ViewBase, abc.ValuesView):
-
     def __contains__(self, value):
         for item in self._impl._items:
             if item[2] == value:
@@ -454,8 +480,11 @@ class _ValuesView(_ViewBase, abc.ValuesView):
         return False
 
     def __iter__(self):
+        return _Iter(len(self), self._iter(self._impl._version))
+
+    def _iter(self, version):
         for item in self._impl._items:
-            if self._version != self._impl._version:
+            if version != self._impl._version:
                 raise RuntimeError("Dictionary changed during iteration")
             yield item[2]
 
@@ -463,12 +492,11 @@ class _ValuesView(_ViewBase, abc.ValuesView):
         lst = []
         for item in self._impl._items:
             lst.append("{!r}".format(item[2]))
-        body = ', '.join(lst)
-        return '{}({})'.format(self.__class__.__name__, body)
+        body = ", ".join(lst)
+        return "{}({})".format(self.__class__.__name__, body)
 
 
 class _KeysView(_ViewBase, abc.KeysView):
-
     def __contains__(self, key):
         for item in self._impl._items:
             if item[1] == key:
@@ -476,8 +504,11 @@ class _KeysView(_ViewBase, abc.KeysView):
         return False
 
     def __iter__(self):
+        return _Iter(len(self), self._iter(self._impl._version))
+
+    def _iter(self, version):
         for item in self._impl._items:
-            if self._version != self._impl._version:
+            if version != self._impl._version:
                 raise RuntimeError("Dictionary changed during iteration")
             yield item[1]
 
@@ -485,5 +516,5 @@ class _KeysView(_ViewBase, abc.KeysView):
         lst = []
         for item in self._impl._items:
             lst.append("{!r}".format(item[1]))
-        body = ', '.join(lst)
-        return '{}({})'.format(self.__class__.__name__, body)
+        body = ", ".join(lst)
+        return "{}({})".format(self.__class__.__name__, body)
